module reverse_conv (in0, in1, in2, out);
	input [7:0] in0; // Module: 193
	input [8:0] in1; // Module: 257
	input [8:0] in2; // Module: 449
	output [24:0] out; // Max value (not included): 22270849

	// Level 1
	wire [15:0] inter_level_1_num_0;
	backward_crt2_193_257 bc_level_1_num_0(in0, in1, inter_level_1_num_0);

	// Level 2
	backward_crt2_449_49601 bc_level_2_num_0(in2, inter_level_1_num_0, out);

endmodule

module backward_crt2_193_257 (in1, in2, out);
	input [7:0] in1;
	input [8:0] in2;
	output reg [15:0] out;
	wire [9:0] inter;
	wire [9:0] inter2;
	wire [8:0] aft;
	wire [8:0] aft2;

	assign inter = 257 + in2 - in1;
	mod_257_513 md1(inter, aft);
	// (193^-1) mod 257 = 4
	assign inter2 = 4*aft[0] + 8*aft[1] + 16*aft[2] + 32*aft[3] + 64*aft[4] + 128*aft[5] + 256*aft[6] + 255*aft[7] + 253*aft[8];
	mod_257_1016 md2(inter2, aft2);

	always @ (*) begin
		out = in1 + aft2*193;
	end

endmodule

module mod_257_513 (in, out);
	input [9:0] in;
	output reg [8:0] out;
	always @ (in)
	begin
		if (in < 9'd257)
		begin
			out = in;
		end
		else
		begin
			out = in - 9'd257;
		end
	end
endmodule

module mod_257_1016 (in, out);
	input [9:0] in;
	output reg [8:0] out;
	always @ (in)
	begin
		if (in < 9'd257)
		begin
			out = in;
		end
		else if (in < 10'd514)
		begin
			out = in - 9'd257;
		end
		else if (in < 10'd771)
		begin
			out = in - 10'd514;
		end
		else
		begin
			out = in - 10'd771;
		end
	end
endmodule


module backward_crt2_449_49601 (in1, in2, out);
	input [8:0] in1;
	input [15:0] in2;
	output reg [24:0] out;
	wire [16:0] inter;
	wire [18:0] inter2;
	wire [15:0] aft;
	wire [15:0] aft2;

	assign inter = 49601 + in2 - in1;
	mod_49601_99201 md1(inter, aft);
	// (449^-1) mod 49601 = 20216
	assign inter2 = 20216*aft[0] + 40432*aft[1] + 31263*aft[2] + 12925*aft[3] + 25850*aft[4] + 2099*aft[5] + 4198*aft[6] + 8396*aft[7] + 16792*aft[8] + 33584*aft[9] + 17567*aft[10] + 35134*aft[11] + 20667*aft[12] + 41334*aft[13] + 33067*aft[14] + 16533*aft[15];
	mod_49601_360057 md2(inter2, aft2);

	always @ (*) begin
		out = in1 + aft2*449;
	end

endmodule

module mod_49601_99201 (in, out);
	input [16:0] in;
	output reg [15:0] out;
	always @ (in)
	begin
		if (in < 16'd49601)
		begin
			out = in;
		end
		else
		begin
			out = in - 16'd49601;
		end
	end
endmodule

module mod_49601_360057 (in, out);
	input [18:0] in;
	output reg [15:0] out;
	always @ (in)
	begin
		if (in < 16'd49601)
		begin
			out = in;
		end
		else if (in < 17'd99202)
		begin
			out = in - 16'd49601;
		end
		else if (in < 18'd148803)
		begin
			out = in - 17'd99202;
		end
		else if (in < 18'd198404)
		begin
			out = in - 18'd148803;
		end
		else if (in < 18'd248005)
		begin
			out = in - 18'd198404;
		end
		else if (in < 19'd297606)
		begin
			out = in - 18'd248005;
		end
		else if (in < 19'd347207)
		begin
			out = in - 19'd297606;
		end
		else
		begin
			out = in - 19'd347207;
		end
	end
endmodule


module atest_bench();
	reg [7:0] in0;  // Module: 193
	reg [8:0] in1;  // Module: 257
	reg [8:0] in2;  // Module: 449
	wire [24:0] out; // Max value (not included): 22270849

	integer l;
	reg dummy;
	integer fori;

	reverse_conv r1 (in0, in1, in2, out);

	initial
	begin
		for (fori = 0; fori < 25'd22270849; fori = fori + 1)
		begin
			in0 = fori%193;
			in1 = fori%257;
			in2 = fori%449;
			#1 dummy = 1;
			$display ("!!! Res=(%d) Expect=(%d)", out, fori);
			l = out;
			if (l != fori)
			begin
				$display ("!!! Error (%d, %d)!!!", fori, l);
			end
			#1 dummy = 1;
		end
	end
endmodule
