module reverse_conv (in0, in1, in2, in3, in4, in5, in6, in7, out);
	input [1:0] in0; // Module: 3
	input [2:0] in1; // Module: 5
	input [2:0] in2; // Module: 7
	input [3:0] in3; // Module: 11
	input [3:0] in4; // Module: 13
	input [4:0] in5; // Module: 17
	input [4:0] in6; // Module: 23
	input [4:0] in7; // Module: 31
	output [27:0] out; // Max value (not included): 181996815

	// Level 1
	wire [3:0] inter_level_1_num_0;
	backward_crt2_3_5 bc_level_1_num_0(in0, in1, inter_level_1_num_0);
	wire [6:0] inter_level_1_num_2;
	backward_crt2_7_11 bc_level_1_num_2(in2, in3, inter_level_1_num_2);
	wire [7:0] inter_level_1_num_4;
	backward_crt2_13_17 bc_level_1_num_4(in4, in5, inter_level_1_num_4);
	wire [9:0] inter_level_1_num_6;
	backward_crt2_23_31 bc_level_1_num_6(in6, in7, inter_level_1_num_6);

	// Level 2
	wire [10:0] inter_level_2_num_0;
	backward_crt2_15_77 bc_level_2_num_0(inter_level_1_num_0, inter_level_1_num_2, inter_level_2_num_0);
	wire [17:0] inter_level_2_num_2;
	backward_crt2_221_713 bc_level_2_num_2(inter_level_1_num_4, inter_level_1_num_6, inter_level_2_num_2);

	// Level 3
	backward_crt2_1155_157573 bc_level_3_num_0(inter_level_2_num_0, inter_level_2_num_2, out);

endmodule

module backward_crt2_3_5 (in1, in2, out);
	input [1:0] in1;
	input [2:0] in2;
	output reg [3:0] out;
	wire [3:0] inter;
	wire [3:0] inter2;
	wire [2:0] aft;
	wire [2:0] aft2;

	assign inter = 5 + in2 - in1;
	mod_5_9 md1(inter, aft);
	// (3^-1) mod 5 = 2
	assign inter2 = 2*aft[0] + 4*aft[1] + 3*aft[2];
	mod_5_9 md2(inter2, aft2);

	always @ (*) begin
		out = in1 + aft2*3;
	end

endmodule

module mod_5_9 (in, out);
	input [3:0] in;
	output reg [2:0] out;
	always @ (in)
	begin
	// we have small max value so we can use table here
	case (in)
		0: out = 0;
		1: out = 1;
		2: out = 2;
		3: out = 3;
		4: out = 4;
		5: out = 0;
		6: out = 1;
		7: out = 2;
		8: out = 3;
		9: out = 4;
		default: out = 0;
	endcase
	end
endmodule

module backward_crt2_7_11 (in1, in2, out);
	input [2:0] in1;
	input [3:0] in2;
	output reg [6:0] out;
	wire [4:0] inter;
	wire [5:0] inter2;
	wire [3:0] aft;
	wire [3:0] aft2;

	assign inter = 11 + in2 - in1;
	mod_11_21 md1(inter, aft);
	// (7^-1) mod 11 = 8
	assign inter2 = 8*aft[0] + 5*aft[1] + 10*aft[2] + 9*aft[3];
	mod_11_32 md2(inter2, aft2);

	always @ (*) begin
		out = in1 + aft2*7;
	end

endmodule

module mod_11_21 (in, out);
	input [4:0] in;
	output reg [3:0] out;
	always @ (in)
	begin
	// we have small max value so we can use table here
	case (in)
		0: out = 0;
		1: out = 1;
		2: out = 2;
		3: out = 3;
		4: out = 4;
		5: out = 5;
		6: out = 6;
		7: out = 7;
		8: out = 8;
		9: out = 9;
		10: out = 10;
		11: out = 0;
		12: out = 1;
		13: out = 2;
		14: out = 3;
		15: out = 4;
		16: out = 5;
		17: out = 6;
		18: out = 7;
		19: out = 8;
		20: out = 9;
		21: out = 10;
		default: out = 0;
	endcase
	end
endmodule

module mod_11_32 (in, out);
	input [5:0] in;
	output reg [3:0] out;
	always @ (in)
	begin
	// we have small max value so we can use table here
	case (in)
		0: out = 0;
		1: out = 1;
		2: out = 2;
		3: out = 3;
		4: out = 4;
		5: out = 5;
		6: out = 6;
		7: out = 7;
		8: out = 8;
		9: out = 9;
		10: out = 10;
		11: out = 0;
		12: out = 1;
		13: out = 2;
		14: out = 3;
		15: out = 4;
		16: out = 5;
		17: out = 6;
		18: out = 7;
		19: out = 8;
		20: out = 9;
		21: out = 10;
		22: out = 0;
		23: out = 1;
		24: out = 2;
		25: out = 3;
		26: out = 4;
		27: out = 5;
		28: out = 6;
		29: out = 7;
		30: out = 8;
		31: out = 9;
		32: out = 10;
		default: out = 0;
	endcase
	end
endmodule

module backward_crt2_13_17 (in1, in2, out);
	input [3:0] in1;
	input [4:0] in2;
	output reg [7:0] out;
	wire [5:0] inter;
	wire [5:0] inter2;
	wire [4:0] aft;
	wire [4:0] aft2;

	assign inter = 17 + in2 - in1;
	mod_17_33 md1(inter, aft);
	// (13^-1) mod 17 = 4
	assign inter2 = 4*aft[0] + 8*aft[1] + 16*aft[2] + 15*aft[3] + 13*aft[4];
	mod_17_56 md2(inter2, aft2);

	always @ (*) begin
		out = in1 + aft2*13;
	end

endmodule

module mod_17_33 (in, out);
	input [5:0] in;
	output reg [4:0] out;
	always @ (in)
	begin
	// we have small max value so we can use table here
	case (in)
		0: out = 0;
		1: out = 1;
		2: out = 2;
		3: out = 3;
		4: out = 4;
		5: out = 5;
		6: out = 6;
		7: out = 7;
		8: out = 8;
		9: out = 9;
		10: out = 10;
		11: out = 11;
		12: out = 12;
		13: out = 13;
		14: out = 14;
		15: out = 15;
		16: out = 16;
		17: out = 0;
		18: out = 1;
		19: out = 2;
		20: out = 3;
		21: out = 4;
		22: out = 5;
		23: out = 6;
		24: out = 7;
		25: out = 8;
		26: out = 9;
		27: out = 10;
		28: out = 11;
		29: out = 12;
		30: out = 13;
		31: out = 14;
		32: out = 15;
		33: out = 16;
		default: out = 0;
	endcase
	end
endmodule

module mod_17_56 (in, out);
	input [5:0] in;
	output reg [4:0] out;
	always @ (in)
	begin
	// we have small max value so we can use table here
	case (in)
		0: out = 0;
		1: out = 1;
		2: out = 2;
		3: out = 3;
		4: out = 4;
		5: out = 5;
		6: out = 6;
		7: out = 7;
		8: out = 8;
		9: out = 9;
		10: out = 10;
		11: out = 11;
		12: out = 12;
		13: out = 13;
		14: out = 14;
		15: out = 15;
		16: out = 16;
		17: out = 0;
		18: out = 1;
		19: out = 2;
		20: out = 3;
		21: out = 4;
		22: out = 5;
		23: out = 6;
		24: out = 7;
		25: out = 8;
		26: out = 9;
		27: out = 10;
		28: out = 11;
		29: out = 12;
		30: out = 13;
		31: out = 14;
		32: out = 15;
		33: out = 16;
		34: out = 0;
		35: out = 1;
		36: out = 2;
		37: out = 3;
		38: out = 4;
		39: out = 5;
		40: out = 6;
		41: out = 7;
		42: out = 8;
		43: out = 9;
		44: out = 10;
		45: out = 11;
		46: out = 12;
		47: out = 13;
		48: out = 14;
		49: out = 15;
		50: out = 16;
		51: out = 0;
		52: out = 1;
		53: out = 2;
		54: out = 3;
		55: out = 4;
		56: out = 5;
		default: out = 0;
	endcase
	end
endmodule

module backward_crt2_23_31 (in1, in2, out);
	input [4:0] in1;
	input [4:0] in2;
	output reg [9:0] out;
	wire [5:0] inter;
	wire [6:0] inter2;
	wire [4:0] aft;
	wire [4:0] aft2;

	assign inter = 31 + in2 - in1;
	mod_31_61 md1(inter, aft);
	// (23^-1) mod 31 = 27
	assign inter2 = 27*aft[0] + 23*aft[1] + 15*aft[2] + 30*aft[3] + 29*aft[4];
	mod_31_124 md2(inter2, aft2);

	always @ (*) begin
		out = in1 + aft2*23;
	end

endmodule

module mod_31_61 (in, out);
	input [5:0] in;
	output reg [4:0] out;
	always @ (in)
	begin
	// we have small max value so we can use table here
	case (in)
		0: out = 0;
		1: out = 1;
		2: out = 2;
		3: out = 3;
		4: out = 4;
		5: out = 5;
		6: out = 6;
		7: out = 7;
		8: out = 8;
		9: out = 9;
		10: out = 10;
		11: out = 11;
		12: out = 12;
		13: out = 13;
		14: out = 14;
		15: out = 15;
		16: out = 16;
		17: out = 17;
		18: out = 18;
		19: out = 19;
		20: out = 20;
		21: out = 21;
		22: out = 22;
		23: out = 23;
		24: out = 24;
		25: out = 25;
		26: out = 26;
		27: out = 27;
		28: out = 28;
		29: out = 29;
		30: out = 30;
		31: out = 0;
		32: out = 1;
		33: out = 2;
		34: out = 3;
		35: out = 4;
		36: out = 5;
		37: out = 6;
		38: out = 7;
		39: out = 8;
		40: out = 9;
		41: out = 10;
		42: out = 11;
		43: out = 12;
		44: out = 13;
		45: out = 14;
		46: out = 15;
		47: out = 16;
		48: out = 17;
		49: out = 18;
		50: out = 19;
		51: out = 20;
		52: out = 21;
		53: out = 22;
		54: out = 23;
		55: out = 24;
		56: out = 25;
		57: out = 26;
		58: out = 27;
		59: out = 28;
		60: out = 29;
		61: out = 30;
		default: out = 0;
	endcase
	end
endmodule

module mod_31_124 (in, out);
	input [6:0] in;
	output reg [4:0] out;
	always @ (in)
	begin
		if (in < 5'd31)
		begin
			out = in;
		end
		else if (in < 6'd62)
		begin
			out = in - 5'd31;
		end
		else if (in < 7'd93)
		begin
			out = in - 6'd62;
		end
		else if (in < 7'd124)
		begin
			out = in - 7'd93;
		end
		else
		begin
			out = in - 7'd124;
		end
	end
endmodule


module backward_crt2_15_77 (in1, in2, out);
	input [3:0] in1;
	input [6:0] in2;
	output reg [10:0] out;
	wire [7:0] inter;
	wire [8:0] inter2;
	wire [6:0] aft;
	wire [6:0] aft2;

	assign inter = 77 + in2 - in1;
	mod_77_153 md1(inter, aft);
	// (15^-1) mod 77 = 36
	assign inter2 = 36*aft[0] + 72*aft[1] + 67*aft[2] + 57*aft[3] + 37*aft[4] + 74*aft[5] + 71*aft[6];
	mod_77_414 md2(inter2, aft2);

	always @ (*) begin
		out = in1 + aft2*15;
	end

endmodule

module mod_77_153 (in, out);
	input [7:0] in;
	output reg [6:0] out;
	always @ (in)
	begin
		if (in < 7'd77)
		begin
			out = in;
		end
		else
		begin
			out = in - 7'd77;
		end
	end
endmodule

module mod_77_414 (in, out);
	input [8:0] in;
	output reg [6:0] out;
	always @ (in)
	begin
		if (in < 7'd77)
		begin
			out = in;
		end
		else if (in < 8'd154)
		begin
			out = in - 7'd77;
		end
		else if (in < 8'd231)
		begin
			out = in - 8'd154;
		end
		else if (in < 9'd308)
		begin
			out = in - 8'd231;
		end
		else if (in < 9'd385)
		begin
			out = in - 9'd308;
		end
		else
		begin
			out = in - 9'd385;
		end
	end
endmodule

module backward_crt2_221_713 (in1, in2, out);
	input [7:0] in1;
	input [9:0] in2;
	output reg [17:0] out;
	wire [10:0] inter;
	wire [12:0] inter2;
	wire [9:0] aft;
	wire [9:0] aft2;

	assign inter = 713 + in2 - in1;
	mod_713_1425 md1(inter, aft);
	// (221^-1) mod 713 = 442
	assign inter2 = 442*aft[0] + 171*aft[1] + 342*aft[2] + 684*aft[3] + 655*aft[4] + 597*aft[5] + 481*aft[6] + 249*aft[7] + 498*aft[8] + 283*aft[9];
	mod_713_4402 md2(inter2, aft2);

	always @ (*) begin
		out = in1 + aft2*221;
	end

endmodule

module mod_713_1425 (in, out);
	input [10:0] in;
	output reg [9:0] out;
	always @ (in)
	begin
		if (in < 10'd713)
		begin
			out = in;
		end
		else
		begin
			out = in - 10'd713;
		end
	end
endmodule

module mod_713_4402 (in, out);
	input [12:0] in;
	output reg [9:0] out;
	always @ (in)
	begin
		if (in < 10'd713)
		begin
			out = in;
		end
		else if (in < 11'd1426)
		begin
			out = in - 10'd713;
		end
		else if (in < 12'd2139)
		begin
			out = in - 11'd1426;
		end
		else if (in < 12'd2852)
		begin
			out = in - 12'd2139;
		end
		else if (in < 12'd3565)
		begin
			out = in - 12'd2852;
		end
		else if (in < 13'd4278)
		begin
			out = in - 12'd3565;
		end
		else
		begin
			out = in - 13'd4278;
		end
	end
endmodule


module backward_crt2_1155_157573 (in1, in2, out);
	input [10:0] in1;
	input [17:0] in2;
	output reg [27:0] out;
	wire [18:0] inter;
	wire [20:0] inter2;
	wire [17:0] aft;
	wire [17:0] aft2;

	assign inter = 157573 + in2 - in1;
	mod_157573_315145 md1(inter, aft);
	// (1155^-1) mod 157573 = 146386
	assign inter2 = 146386*aft[0] + 135199*aft[1] + 112825*aft[2] + 68077*aft[3] + 136154*aft[4] + 114735*aft[5] + 71897*aft[6] + 143794*aft[7] + 130015*aft[8] + 102457*aft[9] + 47341*aft[10] + 94682*aft[11] + 31791*aft[12] + 63582*aft[13] + 127164*aft[14] + 96755*aft[15] + 35937*aft[16] + 71874*aft[17];
	mod_157573_1730665 md2(inter2, aft2);

	always @ (*) begin
		out = in1 + aft2*1155;
	end

endmodule

module mod_157573_315145 (in, out);
	input [18:0] in;
	output reg [17:0] out;
	always @ (in)
	begin
		if (in < 18'd157573)
		begin
			out = in;
		end
		else
		begin
			out = in - 18'd157573;
		end
	end
endmodule

module mod_157573_1730665 (in, out);
	input [20:0] in;
	output reg [17:0] out;
	always @ (in)
	begin
		if (in < 18'd157573)
		begin
			out = in;
		end
		else if (in < 19'd315146)
		begin
			out = in - 18'd157573;
		end
		else if (in < 19'd472719)
		begin
			out = in - 19'd315146;
		end
		else if (in < 20'd630292)
		begin
			out = in - 19'd472719;
		end
		else if (in < 20'd787865)
		begin
			out = in - 20'd630292;
		end
		else if (in < 20'd945438)
		begin
			out = in - 20'd787865;
		end
		else if (in < 21'd1103011)
		begin
			out = in - 20'd945438;
		end
		else if (in < 21'd1260584)
		begin
			out = in - 21'd1103011;
		end
		else if (in < 21'd1418157)
		begin
			out = in - 21'd1260584;
		end
		else if (in < 21'd1575730)
		begin
			out = in - 21'd1418157;
		end
		else
		begin
			out = in - 21'd1575730;
		end
	end
endmodule


module atest_bench();
	reg [1:0] in0;  // Module: 3
	reg [2:0] in1;  // Module: 5
	reg [2:0] in2;  // Module: 7
	reg [3:0] in3;  // Module: 11
	reg [3:0] in4;  // Module: 13
	reg [4:0] in5;  // Module: 17
	reg [4:0] in6;  // Module: 23
	reg [4:0] in7;  // Module: 31
	wire [27:0] out; // Max value (not included): 181996815

	integer l;
	reg dummy;
	integer fori;

	reverse_conv r1 (in0, in1, in2, in3, in4, in5, in6, in7, out);

	initial
	begin
		for (fori = 0; fori < 28'd181996815; fori = fori + 1)
		begin
			in0 = fori%3;
			in1 = fori%5;
			in2 = fori%7;
			in3 = fori%11;
			in4 = fori%13;
			in5 = fori%17;
			in6 = fori%23;
			in7 = fori%31;
			#1 dummy = 1;
			$display ("!!! Res=(%d) Expect=(%d)", out, fori);
			l = out;
			if (l != fori)
			begin
				$display ("!!! Error (%d, %d)!!!", fori, l);
			end
			#1 dummy = 1;
		end
	end
endmodule
