module lut_193_257(out, a, b, clk);
	input [7:0] a;
	input [8:0] b;
	reg [8:0] inter;
	wire [9:0] bmod;
	input clk;
	output wire [8:0] out;

	assign bmod = 4*inter[0] + 8*inter[1] + 16*inter[2] + 32*inter[3] + 64*inter[4] + 128*inter[5] + 256*inter[6] + 255*inter[7] + 253*inter[8];
	mod_257_1016 mod (bmod, out);
	
	always @ (posedge clk) 
	begin 

		if (b >= a)
		begin 
			inter = b - a;
		end 
		else 
		begin 
			inter = 257 + b - a;
		end 
	end
endmodule

module mod_257_1016 (in, out);
	input [9:0] in;
	output reg [8:0] out;
	always @ (in)
	begin
		if (in < 9'd257)
		begin
			out = in;
		end
		else if (in < 10'd514)
		begin
			out = in - 9'd257;
		end
		else if (in < 10'd771)
		begin
			out = in - 10'd514;
		end
		else
		begin
			out = in - 10'd771;
		end
	end
endmodule

module lut_193_449(out, a, b, clk);
	input [7:0] a;
	input [8:0] b;
	reg [8:0] inter;
	wire [10:0] bmod;
	input clk;
	output wire [8:0] out;

	assign bmod = 114*inter[0] + 228*inter[1] + 7*inter[2] + 14*inter[3] + 28*inter[4] + 56*inter[5] + 112*inter[6] + 224*inter[7] + 448*inter[8];
	mod_449_1231 mod (bmod, out);
	
	always @ (posedge clk) 
	begin 

		if (b >= a)
		begin 
			inter = b - a;
		end 
		else 
		begin 
			inter = 449 + b - a;
		end 
	end
endmodule

module mod_449_1231 (in, out);
	input [10:0] in;
	output reg [8:0] out;
	always @ (in)
	begin
		if (in < 9'd449)
		begin
			out = in;
		end
		else if (in < 10'd898)
		begin
			out = in - 9'd449;
		end
		else
		begin
			out = in - 10'd898;
		end
	end
endmodule

module lut_257_449(out, a, b, clk);
	input [8:0] a;
	input [8:0] b;
	reg [8:0] inter;
	wire [11:0] bmod;
	input clk;
	output wire [8:0] out;

	assign bmod = 152*inter[0] + 304*inter[1] + 159*inter[2] + 318*inter[3] + 187*inter[4] + 374*inter[5] + 299*inter[6] + 149*inter[7] + 298*inter[8];
	mod_449_2240 mod (bmod, out);
	
	always @ (posedge clk) 
	begin 

		if (b >= a)
		begin 
			inter = b - a;
		end 
		else 
		begin 
			inter = 449 + b - a;
		end 
	end
endmodule

module mod_449_2240 (in, out);
	input [11:0] in;
	output reg [8:0] out;
	always @ (in)
	begin
		if (in < 9'd449)
		begin
			out = in;
		end
		else if (in < 10'd898)
		begin
			out = in - 9'd449;
		end
		else if (in < 11'd1347)
		begin
			out = in - 10'd898;
		end
		else if (in < 11'd1796)
		begin
			out = in - 11'd1347;
		end
		else
		begin
			out = in - 11'd1796;
		end
	end
endmodule

module mult257(out, data);

	input [8:0] data;
	output reg [15:0] out;
	
	always @ (data) 
	begin
		out <= 193*data;
	end
endmodule

module mult449(out, data);

	input [8:0] data;
	output reg [24:0] out;
	
	always @ (data) 
	begin
		out <= 49601*data;
	end
endmodule

module register(res, x, clk);
	parameter WIDTH=3;
	input  clk;
	input  [WIDTH-1:0] x;
	output reg [WIDTH-1:0] res;

	always @ (posedge clk)
	begin
		res <= x;
	end
endmodule

module adder(result, dataa, datab, clk);
	parameter WIDTH_A=3;
	parameter WIDTH_B=8;
	parameter WIDTH_C=8;
	input clk;
	input  [WIDTH_A - 1:0] dataa;
	input  [WIDTH_B - 1:0] datab;
	output reg [WIDTH_C:0] result;

	always @ (posedge clk) 
	begin 
		result <= dataa + datab;
	end
endmodule

module mix_rdx (res1, res2, res3, x1, x2, x3, clk); 

	input clk;
	input	[7:0] x1;
	input	[8:0] x2;
	input	[8:0] x3;

	output	[7:0] res1;
	output	[8:0] res2;
	output	[8:0] res3;
	
	wire [8:0] out_193_449;

	assign res1 = x1;

	lut_193_257 l193_257 (res2, x1, x2, clk);
	lut_193_449 l193_449 (out_193_449, x1, x3, clk);
	lut_257_449 l257_449 (res3, res2, out_193_449, clk);
endmodule

module par_adder(result, x1, x2, x3, clk);

	input clk;
	input [7:0] x1;
	input [15:0] x2;
	input [24:0] x3;
	output [24:0] result;
	wire [7:0] reg1out;
	wire [16:0] add1out;

	register #(.WIDTH(8)) reg1 (reg1out, x1, clk);
	adder #(.WIDTH_A(8), .WIDTH_B(16), .WIDTH_C(16)) add1 (add1out, reg1out, x2, clk);
	adder #(.WIDTH_A(17), .WIDTH_B(25), .WIDTH_C(24)) add2 (result, add1out, x3, clk);
endmodule

module reverse_conv(result, x1, x2, x3, clk);

	input clk;
	input [7:0] x1;
	input [8:0] x2;
	input [8:0] x3;
	output [24:0] result;

	wire [7:0] mix_rdx1out;
	wire [8:0] mix_rdx2out;
	wire [8:0] mix_rdx3out;

	wire [15:0] mult257out;
	wire [24:0] mult449out;
	mix_rdx mix1 (mix_rdx1out, mix_rdx2out, mix_rdx3out, x1, x2, x3, clk);
	mult257 m257 (mult257out, mix_rdx2out);
	mult449 m449 (mult449out, mix_rdx3out);
	par_adder pa (result, mix_rdx1out, mult257out, mult449out, clk);

endmodule

module atest_bench();
	reg clk;
	reg [7:0] in0;  // Module: 193
	reg [8:0] in1;  // Module: 257
	reg [8:0] in2;  // Module: 449
	wire [24:0] out; // Max value (not included): 22270849

	integer l;
	reg dummy;
	integer i0, prev;

	reverse_conv r1 (out, in0, in1, in2, clk);

	// One test. Ticks: 3
	initial
	begin
		i0 = 15886742;
		clk = 0;
		#1 dummy = 1;
		in0 = i0%193;
		in1 = i0%257;
		in2 = i0%449;
		#1 dummy = 1;
		clk = 1;
		#1 dummy = 1;
		clk = 0;
		#1 dummy = 1;
		clk = 1;
		#1 dummy = 1;
		clk = 0;
		#1 dummy = 1;
		clk = 1;
		#1 dummy = 1;
		clk = 0;
		#1 dummy = 1;
		clk = 1;
		$display ("!!! Res=(%d) Expect=(%d)", out, i0);
		l = out;
		if (l != i0)
		begin
			$display ("!!! Error (%d, %d)!!!", i0, l);
		end
	end
endmodule
