module reverse_conv (clk, in0, in1, in2, out);
	input clk;
	input [1:0] in0; // Module: 3
	input [2:0] in1; // Module: 5
	input [2:0] in2; // Module: 7
	output [6:0] out; // Max value (not included): 105

	// Level 1
	wire [3:0] inter_level_1_num_0;
	backward_crt2_3_5 bc_level_1_num_0(clk, in0, in1, inter_level_1_num_0);
	wire [2:0] in2_prop;
	prop_mod_7 bc_level_1_num_2(clk, in2, in2_prop);

	// Level 2
	backward_crt2_7_15 bc_level_2_num_0(clk, in2_prop, inter_level_1_num_0, out);

endmodule

module backward_crt2_3_5 (clk, in1, in2, out);
	input clk;
	input [1:0] in1;
	input [2:0] in2;
	output reg [3:0] out;
	wire [3:0] inter;
	wire [3:0] inter2;
	wire [2:0] aft;
	wire [2:0] aft2;

	assign inter = 5 + in2 - in1;
	mod_5_9 md1(inter, aft);
	// (3^-1) mod 5 = 2
	assign inter2 = 2*aft[0] + 4*aft[1] + 3*aft[2];
	mod_5_9 md2(inter2, aft2);

	always @ (posedge clk) begin
		out = in1 + aft2*3;
	end

endmodule

module mod_5_9 (in, out);
	input [3:0] in;
	output reg [2:0] out;
	always @ (in)
	begin
	// we have small max value so we can use table here
	case (in)
		0: out = 0;
		1: out = 1;
		2: out = 2;
		3: out = 3;
		4: out = 4;
		5: out = 0;
		6: out = 1;
		7: out = 2;
		8: out = 3;
		9: out = 4;
		default: out = 0;
	endcase
	end
endmodule

module prop_mod_7 (clk, in, out);
	input clk;
	input [2:0] in;
	output reg [2:0] out;

	always @ (posedge clk) begin
		out = in;
	end
endmodule


module backward_crt2_7_15 (clk, in1, in2, out);
	input clk;
	input [2:0] in1;
	input [3:0] in2;
	output reg [6:0] out;
	wire [4:0] inter;
	wire [5:0] inter2;
	wire [3:0] aft;
	wire [3:0] aft2;

	assign inter = 15 + in2 - in1;
	mod_15_29 md1(inter, aft);
	// (7^-1) mod 15 = 13
	assign inter2 = 13*aft[0] + 11*aft[1] + 7*aft[2] + 14*aft[3];
	mod_15_45 md2(inter2, aft2);

	always @ (posedge clk) begin
		out = in1 + aft2*7;
	end

endmodule

module mod_15_29 (in, out);
	input [4:0] in;
	output reg [3:0] out;
	always @ (in)
	begin
	// we have small max value so we can use table here
	case (in)
		0: out = 0;
		1: out = 1;
		2: out = 2;
		3: out = 3;
		4: out = 4;
		5: out = 5;
		6: out = 6;
		7: out = 7;
		8: out = 8;
		9: out = 9;
		10: out = 10;
		11: out = 11;
		12: out = 12;
		13: out = 13;
		14: out = 14;
		15: out = 0;
		16: out = 1;
		17: out = 2;
		18: out = 3;
		19: out = 4;
		20: out = 5;
		21: out = 6;
		22: out = 7;
		23: out = 8;
		24: out = 9;
		25: out = 10;
		26: out = 11;
		27: out = 12;
		28: out = 13;
		29: out = 14;
		default: out = 0;
	endcase
	end
endmodule

module mod_15_45 (in, out);
	input [5:0] in;
	output reg [3:0] out;
	always @ (in)
	begin
	// we have small max value so we can use table here
	case (in)
		0: out = 0;
		1: out = 1;
		2: out = 2;
		3: out = 3;
		4: out = 4;
		5: out = 5;
		6: out = 6;
		7: out = 7;
		8: out = 8;
		9: out = 9;
		10: out = 10;
		11: out = 11;
		12: out = 12;
		13: out = 13;
		14: out = 14;
		15: out = 0;
		16: out = 1;
		17: out = 2;
		18: out = 3;
		19: out = 4;
		20: out = 5;
		21: out = 6;
		22: out = 7;
		23: out = 8;
		24: out = 9;
		25: out = 10;
		26: out = 11;
		27: out = 12;
		28: out = 13;
		29: out = 14;
		30: out = 0;
		31: out = 1;
		32: out = 2;
		33: out = 3;
		34: out = 4;
		35: out = 5;
		36: out = 6;
		37: out = 7;
		38: out = 8;
		39: out = 9;
		40: out = 10;
		41: out = 11;
		42: out = 12;
		43: out = 13;
		44: out = 14;
		45: out = 0;
		default: out = 0;
	endcase
	end
endmodule


module atest_bench();
	reg clk;
	reg [1:0] in0;  // Module: 3
	reg [2:0] in1;  // Module: 5
	reg [2:0] in2;  // Module: 7
	wire [6:0] out; // Max value (not included): 105

	integer l;
	reg dummy;
	integer i0;
	integer i1;

	reverse_conv r1 (clk, in0, in1, in2, out);

	// One test. Ticks: 2 [Data input] + 2 [Data output]
	initial
	begin
		i0 = 104;
		clk = 0;
		#1 dummy = 1;
		in0 = i0%3;
		in1 = i0%5;
		in2 = i0%7;
		#1 dummy = 1;
		clk = 1;
		#1 dummy = 1;
		i1 = 40;
		clk = 0;
		#1 dummy = 1;
		in0 = i1%3;
		in1 = i1%5;
		in2 = i1%7;
		#1 dummy = 1;
		clk = 1;
		#1 dummy = 1;
		$display ("!!! Res=(%d) Expect=(%d)", out, i0);
		l = out;
		if (l != i0)
		begin
			$display ("!!! Error (%d, %d)!!!", i0, l);
		end
		#1 dummy = 1;
		clk = 0;
		#1 dummy = 1;
		clk = 1;
		#1 dummy = 1;
		$display ("!!! Res=(%d) Expect=(%d)", out, i1);
		l = out;
		if (l != i1)
		begin
			$display ("!!! Error (%d, %d)!!!", i1, l);
		end
		#1 dummy = 1;
		clk = 0;
		#1 dummy = 1;
		clk = 1;
		#1 dummy = 1;
	end
endmodule
