module reverse_conv (clk, in0, in1, in2, out);
	input clk;
	input [8:0] in0; // Module: 511
	input [8:0] in1; // Module: 512
	input [9:0] in2; // Module: 513
	output [26:0] out; // Max value (not included): 134217216

	// Level 1
	wire [17:0] inter_level_1_num_0;
	backward_crt2_511_512 bc_level_1_num_0(clk, in0, in1, inter_level_1_num_0);
	wire [9:0] in2_prop;
	prop_mod_513 bc_level_1_num_2(clk, in2, in2_prop);

	// Level 2
	backward_crt2_513_261632 bc_level_2_num_0(clk, in2_prop, inter_level_1_num_0, out);

endmodule

module backward_crt2_511_512 (clk, in1, in2, out);
	input clk;
	input [8:0] in1;
	input [8:0] in2;
	output reg [17:0] out;
	wire [9:0] inter;
	wire [12:0] inter2;
	wire [8:0] aft;
	wire [8:0] aft2;

	assign inter = 512 + in2 - in1;
	mod_512_1023 md1(inter, aft);
	// (511^-1) mod 512 = 511
	assign inter2 = 511*aft[0] + 510*aft[1] + 508*aft[2] + 504*aft[3] + 496*aft[4] + 480*aft[5] + 448*aft[6] + 384*aft[7] + 256*aft[8];
	mod_512_4097 md2(inter2, aft2);

	always @ (posedge clk) begin
		out = in1 + aft2*511;
	end

endmodule

module mod_512_1023 (in, out);
	input [9:0] in;
	output reg [8:0] out;
	always @ (in)
	begin
		if (in < 10'd512)
		begin
			out = in;
		end
		else
		begin
			out = in - 10'd512;
		end
	end
endmodule

module mod_512_4097 (in, out);
	input [12:0] in;
	output reg [8:0] out;
	always @ (in)
	begin
		if (in < 10'd512)
		begin
			out = in;
		end
		else if (in < 11'd1024)
		begin
			out = in - 10'd512;
		end
		else if (in < 11'd1536)
		begin
			out = in - 11'd1024;
		end
		else if (in < 12'd2048)
		begin
			out = in - 11'd1536;
		end
		else if (in < 12'd2560)
		begin
			out = in - 12'd2048;
		end
		else if (in < 12'd3072)
		begin
			out = in - 12'd2560;
		end
		else if (in < 12'd3584)
		begin
			out = in - 12'd3072;
		end
		else if (in < 13'd4096)
		begin
			out = in - 12'd3584;
		end
		else
		begin
			out = in - 13'd4096;
		end
	end
endmodule

module prop_mod_513 (clk, in, out);
	input clk;
	input [9:0] in;
	output reg [9:0] out;

	always @ (posedge clk) begin
		out = in;
	end
endmodule


module backward_crt2_513_261632 (clk, in1, in2, out);
	input clk;
	input [9:0] in1;
	input [17:0] in2;
	output reg [26:0] out;
	wire [18:0] inter;
	wire [21:0] inter2;
	wire [17:0] aft;
	wire [17:0] aft2;

	assign inter = 261632 + in2 - in1;
	mod_261632_523263 md1(inter, aft);
	// (513^-1) mod 261632 = 130561
	assign inter2 = 130561*aft[0] + 261122*aft[1] + 260612*aft[2] + 259592*aft[3] + 257552*aft[4] + 253472*aft[5] + 245312*aft[6] + 228992*aft[7] + 196352*aft[8] + 131072*aft[9] + 512*aft[10] + 1024*aft[11] + 2048*aft[12] + 4096*aft[13] + 8192*aft[14] + 16384*aft[15] + 32768*aft[16] + 65536*aft[17];
	mod_261632_2355199 md2(inter2, aft2);

	always @ (posedge clk) begin
		out = in1 + aft2*513;
	end

endmodule

module mod_261632_523263 (in, out);
	input [18:0] in;
	output reg [17:0] out;
	always @ (in)
	begin
		if (in < 18'd261632)
		begin
			out = in;
		end
		else
		begin
			out = in - 18'd261632;
		end
	end
endmodule

module mod_261632_2355199 (in, out);
	input [21:0] in;
	output reg [17:0] out;
	always @ (in)
	begin
		if (in < 18'd261632)
		begin
			out = in;
		end
		else if (in < 19'd523264)
		begin
			out = in - 18'd261632;
		end
		else if (in < 20'd784896)
		begin
			out = in - 19'd523264;
		end
		else if (in < 20'd1046528)
		begin
			out = in - 20'd784896;
		end
		else if (in < 21'd1308160)
		begin
			out = in - 20'd1046528;
		end
		else if (in < 21'd1569792)
		begin
			out = in - 21'd1308160;
		end
		else if (in < 21'd1831424)
		begin
			out = in - 21'd1569792;
		end
		else if (in < 21'd2093056)
		begin
			out = in - 21'd1831424;
		end
		else if (in < 22'd2354688)
		begin
			out = in - 21'd2093056;
		end
		else
		begin
			out = in - 22'd2354688;
		end
	end
endmodule


module atest_bench();
	reg clk;
	reg [8:0] in0;  // Module: 511
	reg [8:0] in1;  // Module: 512
	reg [9:0] in2;  // Module: 513
	wire [26:0] out; // Max value (not included): 134217216

	integer l;
	reg dummy;
	integer i0;
	integer i1;

	reverse_conv r1 (clk, in0, in1, in2, out);

	// One test. Ticks: 2 [Data input] + 2 [Data output]
	initial
	begin
		i0 = 30925514;
		clk = 0;
		#1 dummy = 1;
		in0 = i0%511;
		in1 = i0%512;
		in2 = i0%513;
		#1 dummy = 1;
		clk = 1;
		#1 dummy = 1;
		i1 = 105092017;
		clk = 0;
		#1 dummy = 1;
		in0 = i1%511;
		in1 = i1%512;
		in2 = i1%513;
		#1 dummy = 1;
		clk = 1;
		#1 dummy = 1;
		$display ("!!! Res=(%d) Expect=(%d)", out, i0);
		l = out;
		if (l != i0)
		begin
			$display ("!!! Error (%d, %d)!!!", i0, l);
		end
		#1 dummy = 1;
		clk = 0;
		#1 dummy = 1;
		clk = 1;
		#1 dummy = 1;
		$display ("!!! Res=(%d) Expect=(%d)", out, i1);
		l = out;
		if (l != i1)
		begin
			$display ("!!! Error (%d, %d)!!!", i1, l);
		end
		#1 dummy = 1;
		clk = 0;
		#1 dummy = 1;
		clk = 1;
		#1 dummy = 1;
	end
endmodule
